# Copyright 2020 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifer: Apache-2.0

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import csv
import os
import collections
import numpy as np
import pandas as pd
import tensorflow as tf
from itertools import product
import matplotlib.pyplot as plt
from tensorflow.python.saved_model import tag_constants
from sklearn.metrics import precision_score, recall_score, f1_score, accuracy_score, confusion_matrix

from bert import modeling
from bert import optimization
from bert import tokenization

tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
            

class InputExample(object):
    """A single training/test example for simple sequence classification."""

    def __init__(self, guid, text_a, text_b=None, label=None):
        """Constructs a InputExample.

        Args:
          guid: Unique id for the example.
          text_a: string. The untokenized text of the first sequence. For single
            sequence tasks, only this sequence must be specified.
          text_b: (Optional) string. The untokenized text of the second sequence.
            Only must be specified for sequence pair tasks.
          label: (Optional) string. The label of the example. This should be
            specified for train and dev examples, but not for test examples.
        """
        self.guid = guid
        self.text_a = text_a
        self.text_b = text_b
        self.label = label


class InputFeatures(object):
    """A single set of features of data."""

    def __init__(self,
               input_ids,
               input_mask,
               segment_ids,
               label_id,
               is_real_example=True):
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids
        self.label_id = label_id
        self.is_real_example = is_real_example


class TextProcessor():
    """Processor for the data set."""

    def _read_tsv(cls, input_file, quotechar=None):
        """Reads a csv separated value file."""
        df = pd.read_csv(input_file).astype(str).values.tolist()
        return df

    def get_train_examples(self, data_dir):
        """read train.csv."""
        return self._create_examples(
            self._read_tsv(os.path.join(data_dir, "train.csv")), "train")

    def get_dev_examples(self, data_dir):
        """read dev.csv."""
        return self._create_examples(
            self._read_tsv(os.path.join(data_dir, "dev.csv")), "dev")

    def get_test_examples(self, data_dir):
        """read test.csv."""
        return self._create_examples(
            self._read_tsv(os.path.join(data_dir, "test.csv")), "test")

    def get_labels(self):
        """default_labels."""
        return ["0", "1"]

    def _create_examples(self, lines, set_type):
        """Creates examples for the training and dev sets.
           Assumes that,
           2nd column (index: 1) in test.csv is text input.
           2nd column (index: 1) in train.csv is text input.
           3rd column (index: 2) in train.csv is class label.
        """

        examples = []
        for (i, line) in enumerate(lines):
            # Train, dev and test set has a header
            guid = "%s-%s" % (set_type, i)
            if set_type == "test":
                text_a = tokenization.convert_to_unicode(line[1])
                label = "0"
            else:

                text_a = tokenization.convert_to_unicode(line[1])
                label = tokenization.convert_to_unicode(line[2])
            examples.append(
              InputExample(guid=guid, text_a=text_a, text_b=None, label=label))
        return examples


class BertClassification(object):
    
    """
    configs: CONFIGS object with configuration params
    """
    
    def __init__(self, configs = None):
        
        self.Configs = configs
        self.tokenizer = tokenization.FullTokenizer(vocab_file=self.Configs.vocab_file, do_lower_case=self.Configs.do_lower_case)
        
        bert_config = modeling.BertConfig.from_json_file(self.Configs.bert_config_file)
        
        if self.Configs.max_seq_length > bert_config.max_position_embeddings:
            raise ValueError(
            "Cannot use sequence length %d because the BERT model "
            "was only trained up to sequence length %d" %
            (self.Configs.max_seq_length, bert_config.max_position_embeddings))

        tf.io.gfile.makedirs(self.Configs.output_dir)
        self.processor = TextProcessor()

        tpu_cluster_resolver = None
        if self.Configs.use_tpu and self.Configs.tpu_name:
            tpu_cluster_resolver = tf.contrib.cluster_resolver.TPUClusterResolver(
            self.Configs.tpu_name, zone=self.Configs.tpu_zone, project=self.Configs.gcp_project)

        is_per_host = tf.contrib.tpu.InputPipelineConfig.PER_HOST_V2
        run_config = tf.contrib.tpu.RunConfig(
            cluster=tpu_cluster_resolver,
            master=self.Configs.master,
            model_dir=self.Configs.output_dir,
            save_checkpoints_steps=self.Configs.save_checkpoints_steps,
            tpu_config=tf.contrib.tpu.TPUConfig(
              iterations_per_loop=self.Configs.iterations_per_loop,
              num_shards=self.Configs.num_tpu_cores,
              per_host_input_for_training=is_per_host))

        self.train_examples = None
        self.num_train_steps = None
        self.num_warmup_steps = None
        if self.Configs.do_train:
            self.train_examples = self.processor.get_train_examples(self.Configs.data_dir)
        self.num_train_steps = int(
            len(self.train_examples) / self.Configs.train_batch_size * self.Configs.num_train_epochs)
        self.num_warmup_steps = int(self.num_train_steps * self.Configs.warmup_proportion)

        model_fn = self.model_fn_builder(
          bert_config=bert_config,
          num_labels=len(self.Configs.label_list),
          init_checkpoint=self.Configs.init_checkpoint,
          learning_rate=self.Configs.learning_rate,
          num_train_steps=self.num_train_steps,
          num_warmup_steps=self.num_warmup_steps,
          use_tpu=self.Configs.use_tpu,
          use_one_hot_embeddings=self.Configs.use_tpu)

        self.estimator = tf.contrib.tpu.TPUEstimator(
          use_tpu=self.Configs.use_tpu,
          model_fn=model_fn,
          config=run_config,
          train_batch_size=self.Configs.train_batch_size,
          eval_batch_size=self.Configs.eval_batch_size,
          predict_batch_size=self.Configs.predict_batch_size)

    def convert_single_example(self, example, label_list, max_seq_length,
                               tokenizer):
        """Converts a single `InputExample` into a single `InputFeatures`."""

        label_map = {}
        for (i, label) in enumerate(label_list):
            label_map[label] = i

        tokens_a = tokenizer.tokenize(example.text_a)


        if len(tokens_a) > max_seq_length - 2:
            tokens_a = tokens_a[0:(max_seq_length - 2)]

        tokens = ["[CLS]"] + tokens_a + ['[SEP]']
        input_ids = tokenizer.convert_tokens_to_ids(tokens)
        input_mask = [1] * len(input_ids)

        while len(input_ids) < max_seq_length:
            input_ids.append(0)
            input_mask.append(0)

        segment_ids = [0] * len(input_ids)

        assert len(input_ids) == max_seq_length
        assert len(input_mask) == max_seq_length
        assert len(segment_ids) == max_seq_length

        label_id = label_map[example.label]

        feature = InputFeatures(
          input_ids=input_ids,
          input_mask=input_mask,
          segment_ids=segment_ids,
          label_id=label_id,
          is_real_example=True)
        return feature

    def convert_test_single_example(self, text, max_seq_length, tokenizer):
        
        """Converts a single test `InputExample` into a single `InputFeatures`."""

        tokens_a = tokenizer.tokenize(text)

        if len(tokens_a) > max_seq_length - 2:
            tokens_a = tokens_a[0:(max_seq_length - 2)]

        tokens = ["[CLS]"] + tokens_a + ['[SEP]']
        input_ids = tokenizer.convert_tokens_to_ids(tokens)
        input_mask = [1] * len(input_ids)

        while len(input_ids) < max_seq_length:
            input_ids.append(0)
            input_mask.append(0)
        segment_ids = [0] * len(input_ids)

        feature = InputFeatures(
          input_ids=input_ids,
          input_mask=input_mask,
          segment_ids=segment_ids,
          label_id=0
        )
        return feature

    def file_based_convert_examples_to_features(
        self, examples, label_list, max_seq_length, tokenizer, output_file):
        """Convert a set of `InputExample`s to a TFRecord file."""

        writer = tf.python_io.TFRecordWriter(output_file)

        for (ex_index, example) in enumerate(examples):
            if ex_index % 10000 == 0:
                tf.logging.info("Writing example %d of %d" % (ex_index, len(examples)))

            feature = self.convert_single_example(example, label_list,
                                         max_seq_length, tokenizer)

            def create_int_feature(values):
                f = tf.train.Feature(int64_list=tf.train.Int64List(value=list(values)))
                return f

            features = collections.OrderedDict()
            features["input_ids"] = create_int_feature(feature.input_ids)
            features["input_mask"] = create_int_feature(feature.input_mask)
            features["segment_ids"] = create_int_feature(feature.segment_ids)
            features["label_ids"] = create_int_feature([feature.label_id])
            features["is_real_example"] = create_int_feature(
                [int(feature.is_real_example)])

            tf_example = tf.train.Example(features=tf.train.Features(feature=features))
            writer.write(tf_example.SerializeToString())
        writer.close()

    def file_based_input_fn_builder(self, input_file, seq_length, is_training,
                                    drop_remainder):
        """Creates an `input_fn` closure to be passed to TPUEstimator."""

        name_to_features = {
          "input_ids": tf.FixedLenFeature([seq_length], tf.int64),
          "input_mask": tf.FixedLenFeature([seq_length], tf.int64),
          "segment_ids": tf.FixedLenFeature([seq_length], tf.int64),
          "label_ids": tf.FixedLenFeature([], tf.int64),
          "is_real_example": tf.FixedLenFeature([], tf.int64),
        }

        def _decode_record(record, name_to_features):
            """Decodes a record to a TensorFlow example."""
            example = tf.parse_single_example(record, name_to_features)

            # tf.Example only supports tf.int64, but the TPU only supports tf.int32.
            # So cast all int64 to int32.
            for name in list(example.keys()):
                t = example[name]
                if t.dtype == tf.int64:
                    t = tf.to_int32(t)
                example[name] = t
            return example

        def input_fn(params):
            """The actual input function."""
            batch_size = params["batch_size"]

            # For training, we want a lot of parallel reading and shuffling.
            # For eval, we want no shuffling and parallel reading doesn't matter.
            d = tf.data.TFRecordDataset(input_file)
            if is_training:
                d = d.repeat()
                d = d.shuffle(buffer_size=100)

            d = d.apply(
                tf.contrib.data.map_and_batch(
                    lambda record: _decode_record(record, name_to_features),
                    batch_size=batch_size,
                    drop_remainder=drop_remainder))
            return d

        return input_fn

    def create_model(self, bert_config, is_training, input_ids, input_mask, segment_ids,
                     labels, num_labels, use_one_hot_embeddings):
        """Creates a classification model."""
        
        model = modeling.BertModel(
          config=bert_config,
          is_training=is_training,
          input_ids=input_ids,
          input_mask=input_mask,
          token_type_ids=segment_ids,
          use_one_hot_embeddings=use_one_hot_embeddings)

        output_layer = model.get_pooled_output()

        hidden_size = output_layer.shape[-1].value

        output_weights = tf.get_variable(
          "output_weights", [num_labels, hidden_size],
          initializer=tf.truncated_normal_initializer(stddev=0.02))

        output_bias = tf.get_variable(
          "output_bias", [num_labels], initializer=tf.zeros_initializer())

        with tf.variable_scope("loss"):
            if is_training:
                output_layer = tf.nn.dropout(output_layer, keep_prob=0.9)

        logits = tf.matmul(output_layer, output_weights, transpose_b=True)
        logits = tf.nn.bias_add(logits, output_bias)
        probabilities = tf.nn.softmax(logits, axis=-1)
        log_probs = tf.nn.log_softmax(logits, axis=-1)

        one_hot_labels = tf.one_hot(labels, depth=num_labels, dtype=tf.float32)

        per_example_loss = -tf.reduce_sum(one_hot_labels * log_probs, axis=-1)
        loss = tf.reduce_mean(per_example_loss)

        return (loss, per_example_loss, logits, probabilities)

    def model_fn_builder(self, bert_config, num_labels, init_checkpoint, learning_rate,
                         num_train_steps, num_warmup_steps, use_tpu,
                         use_one_hot_embeddings):
        """Returns `model_fn` closure for TPUEstimator."""

        def model_fn(features, labels, mode, params):  # pylint: disable=unused-argument
            """The `model_fn` for TPUEstimator."""

            tf.logging.info("*** Features ***")
            for name in sorted(features.keys()):
                tf.logging.info("  name = %s, shape = %s" % (name, features[name].shape))

            input_ids = features["input_ids"]
            input_mask = features["input_mask"]
            segment_ids = features["segment_ids"]
            label_ids = features["label_ids"]
            is_real_example = None
            
            if "is_real_example" in features:
                is_real_example = tf.cast(features["is_real_example"], dtype=tf.float32)
            else:
                is_real_example = tf.ones(tf.shape(label_ids), dtype=tf.float32)

            is_training = (mode == tf.estimator.ModeKeys.TRAIN)

            (total_loss, per_example_loss, logits, probabilities) = self.create_model(
                bert_config, is_training, input_ids, input_mask, segment_ids, label_ids,
                num_labels, use_one_hot_embeddings)

            tvars = tf.trainable_variables()
            initialized_variable_names = {}
            scaffold_fn = None
            if init_checkpoint:
                (assignment_map, initialized_variable_names
                ) = modeling.get_assignment_map_from_checkpoint(tvars, init_checkpoint)
                if use_tpu:

                    def tpu_scaffold():
                        tf.train.init_from_checkpoint(init_checkpoint, assignment_map)
                        return tf.train.Scaffold()

                    scaffold_fn = tpu_scaffold
                else:
                    tf.train.init_from_checkpoint(init_checkpoint, assignment_map)

            tf.logging.info("**** Trainable Variables ****")
            for var in tvars:
                init_string = ""
                if var.name in initialized_variable_names:
                    init_string = ", *INIT_FROM_CKPT*"
                tf.logging.info("  name = %s, shape = %s%s", var.name, var.shape,
                              init_string)

            output_spec = None
            if mode == tf.estimator.ModeKeys.TRAIN:

                train_op = optimization.create_optimizer(
                  total_loss, learning_rate, num_train_steps, num_warmup_steps, use_tpu)

                output_spec = tf.contrib.tpu.TPUEstimatorSpec(
                  mode=mode,
                  loss=total_loss,
                  train_op=train_op,
                  scaffold_fn=scaffold_fn)
            elif mode == tf.estimator.ModeKeys.EVAL:

                def metric_fn(per_example_loss, label_ids, logits, is_real_example):
                    predictions = tf.argmax(logits, axis=-1, output_type=tf.int32)
                    accuracy = tf.compat.v1.metrics.accuracy(
                        labels=label_ids, predictions=predictions, weights=is_real_example)
                    loss = tf.compat.v1.metrics.mean(values=per_example_loss, weights=is_real_example)
                    return {
                        "eval_accuracy": accuracy,
                        "eval_loss": loss,
                    }

                eval_metrics = (metric_fn,
                              [per_example_loss, label_ids, logits, is_real_example])
                output_spec = tf.contrib.tpu.TPUEstimatorSpec(
                  mode=mode,
                  loss=total_loss,
                  eval_metrics=eval_metrics,
                  scaffold_fn=scaffold_fn)
            else:
                output_spec = tf.contrib.tpu.TPUEstimatorSpec(
                  mode=mode,
                  predictions={"probabilities": probabilities},
                  scaffold_fn=scaffold_fn)
            return output_spec

        return model_fn

    def train(self):
        """Finetunes the classifier according to the given training data."""

        tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.INFO)

        tokenization.validate_case_matches_checkpoint(self.Configs.do_lower_case,
                                                    self.Configs.init_checkpoint)

        if not self.Configs.do_train and not self.Configs.do_eval:
            raise ValueError(
            "At least one of `do_train` or `do_eval`.")

        if self.Configs.do_train:
            train_file = os.path.join(self.Configs.output_dir, "train.tf_record")
            self.file_based_convert_examples_to_features(
                self.train_examples, self.Configs.label_list, self.Configs.max_seq_length, 
                self.tokenizer, train_file)
            
            tf.logging.info("***** Running training *****")
            tf.logging.info("  Num examples = %d", len(self.train_examples))
            tf.logging.info("  Batch size = %d", self.Configs.train_batch_size)
            tf.logging.info("  Num steps = %d", self.num_train_steps)
            train_input_fn = self.file_based_input_fn_builder(
                input_file=train_file,
                seq_length=self.Configs.max_seq_length,
                is_training=True,
                drop_remainder=True)
            self.estimator.train(input_fn=train_input_fn, max_steps=self.num_train_steps)

        if self.Configs.do_eval:
            eval_examples = self.processor.get_dev_examples(self.Configs.data_dir)
            num_actual_eval_examples = len(eval_examples)

            eval_file = os.path.join(self.Configs.output_dir, "eval.tf_record")
            self.file_based_convert_examples_to_features(
                eval_examples, self.Configs.label_list, self.Configs.max_seq_length, 
                self.tokenizer, eval_file)

            tf.logging.info("***** Running evaluation *****")
            tf.logging.info("  Num examples = %d (%d actual, %d padding)",
                            len(eval_examples), num_actual_eval_examples,
                            len(eval_examples) - num_actual_eval_examples)
            tf.logging.info("  Batch size = %d", self.Configs.eval_batch_size)

            # This tells the estimator to run through the entire set.
            eval_steps = None

            eval_drop_remainder = True if self.Configs.use_tpu else False
            eval_input_fn = self.file_based_input_fn_builder(
                input_file=eval_file,
                seq_length=self.Configs.max_seq_length,
                is_training=False,
                drop_remainder=eval_drop_remainder)

            result = self.estimator.evaluate(input_fn=eval_input_fn, steps=eval_steps)

            output_eval_file = os.path.join(self.Configs.output_dir, "eval_results.txt")
            with tf.gfile.GFile(output_eval_file, "w") as writer:
                tf.logging.info("***** Eval results *****")
                for key in sorted(result.keys()):
                    tf.logging.info("  %s = %s", key, str(result[key]))
                    writer.write("%s = %s\n" % (key, str(result[key])))
        
    def test(self, to_csv=True):
        """
        Predicts the model according to the given test data. 
        returns predictions of shape (n_samples, n_labels)
        """
        
        predict_examples = self.processor.get_test_examples(self.Configs.data_dir)
        num_actual_predict_examples = len(predict_examples)
        
        predict_file = os.path.join(self.Configs.output_dir, "predict.tf_record")
        self.file_based_convert_examples_to_features(predict_examples, self.Configs.label_list,
                                                self.Configs.max_seq_length, self.tokenizer,
                                                predict_file)

        tf.logging.info("***** Running prediction*****")
        tf.logging.info("  Num examples = %d (%d actual, %d padding)",
                        len(predict_examples), num_actual_predict_examples,
                        len(predict_examples) - num_actual_predict_examples)
        tf.logging.info("  Batch size = %d", self.Configs.predict_batch_size)

        predict_drop_remainder = True if self.Configs.use_tpu else False
        predict_input_fn = self.file_based_input_fn_builder(
            input_file=predict_file,
            seq_length=self.Configs.max_seq_length,
            is_training=False,
            drop_remainder=predict_drop_remainder)
        
        result = self.estimator.predict(input_fn=predict_input_fn)
        
        predictions=[]
        for (i, prediction) in enumerate(result):
            predictions.append(prediction["probabilities"])
        print("Predictions = ",len(predictions))
        if to_csv:    
            output_predict_file = os.path.join(self.Configs.output_dir, "test_results.tsv")
            with tf.gfile.GFile(output_predict_file, "w") as writer:
                num_written_lines = 0
                tf.logging.info("***** Predict results *****saved in*****test_results.tsv")
                for (i, prediction) in enumerate(predictions):
                    probabilities = prediction
                    if i >= num_actual_predict_examples:
                        break
                    output_line = "\t".join(str(class_probability) for class_probability in probabilities) + "\n"
                    writer.write(output_line)
                    num_written_lines += 1
            assert num_written_lines == num_actual_predict_examples
            
        return predictions

    def create_infer_model(self, input_ids, input_mask, segment_ids, labels, num_labels):
        """create inferencing model, returns probabilties"""

        model = modeling.BertModel(
          config=modeling.BertConfig.from_json_file(self.Configs.bert_config_file),
          is_training=False,
          input_ids=input_ids,
          input_mask=input_mask,
          token_type_ids=segment_ids,
          use_one_hot_embeddings=False)

        output_layer = model.get_pooled_output()
        hidden_size = output_layer.shape[-1].value

        output_weights = tf.get_variable(
          "output_weights", [num_labels, hidden_size],
          initializer=tf.truncated_normal_initializer(stddev=0.02))

        output_bias = tf.get_variable(
          "output_bias", [num_labels], initializer=tf.zeros_initializer())

        logits = tf.matmul(output_layer, output_weights, transpose_b=True)
        logits = tf.nn.bias_add(logits, output_bias)
        probabilities = tf.nn.sigmoid(logits)
        return probabilities

    
    def infer_fn_builder(self, num_labels):

        def model_fn(features):
            input_ids = features["input_ids"]
            input_mask = features["input_mask"]
            segment_ids = features["segment_ids"]
            label_ids = features["label_ids"]
            is_real_example = None

            probabilities = self.create_infer_model(
                input_ids, input_mask, segment_ids, label_ids, num_labels)

            output_spec = tf.estimator.EstimatorSpec(
                predictions={"probabilities": probabilities},
                mode='infer'
            )
            return output_spec

        return model_fn

    
    def export_model(self):
        """exports a model checkpoint to .pb format (production/freezed graph and weigh)"""

        def serving_input_fn():
            label_ids = tf.placeholder(tf.int32, [None], name='label_ids')
            input_ids = tf.placeholder(tf.int32, [None, self.Configs.max_seq_length], name='input_ids')
            input_mask = tf.placeholder(tf.int32, [None, self.Configs.max_seq_length], name='input_mask')
            segment_ids = tf.placeholder(tf.int32, [None, self.Configs.max_seq_length], name='segment_ids')
            input_fn = tf.estimator.export.build_raw_serving_input_receiver_fn({
                'label_ids': label_ids,
                'input_ids': input_ids,
                'input_mask': input_mask,
                'segment_ids': segment_ids,
            })()
            return input_fn

        model_fn = self.infer_fn_builder(num_labels=len(self.Configs.label_list))
        estimator = tf.estimator.Estimator(model_fn=model_fn,
                                           config=tf.estimator.RunConfig(model_dir=self.Configs.output_dir))

        estimator._export_to_tpu = False
        export_dir = estimator.export_saved_model(export_dir_base=self.Configs.export_path,
                                                  serving_input_receiver_fn=serving_input_fn)
        self.Configs.export_dir = export_dir.decode("utf-8")

        return export_dir.decode("utf-8")

    def chunks(self, text_list, size):
        """
        creates batches on text samples.
        
        Parameters
        ----------
        text_list : text samples of shape (n_samples,)
        size: batch size
        """
        
        for i in range(0, len(text_list), size):
            yield text_list[i:i+size]

    def inference(self, test_filename, batch_size=64):
        """
        performs inference using freezed (.pb) model.
        
        Parameters
        ----------
        test_filename : input file with format as of test.csv
        batch_size: Number of samples in a batch.
        """

        max_seq_length, tokenizer = self.Configs.max_seq_length, self.tokenizer
        df = pd.read_csv(test_filename)

        graph = tf.Graph()
        with graph.as_default():
            with tf.Session() as sess:

                tf.saved_model.loader.load(sess, [tag_constants.SERVING], self.Configs.export_dir)
                tensor_input_ids = graph.get_tensor_by_name('input_ids_1:0')
                tensor_input_mask = graph.get_tensor_by_name('input_mask_1:0')
                tensor_label_ids = graph.get_tensor_by_name('label_ids_1:0')
                tensor_segment_ids = graph.get_tensor_by_name('segment_ids_1:0')
                tensor_outputs = graph.get_tensor_by_name('Sigmoid:0')

                batches = list(self.chunks(df.text.tolist(),batch_size))

                predictions = []
                for i,batch in enumerate(batches):
                    print('Processing batch no. ',i)
                    features = [self.convert_test_single_example(txt, max_seq_length, tokenizer) for txt in batch]
                    input_ids = np.array([feature.input_ids for feature in features]).reshape(-1, max_seq_length)
                    input_mask = np.array([feature.input_mask for feature in features]).reshape(-1, max_seq_length)
                    label_ids = np.array([feature.label_id for feature in features])
                    segment_ids = np.array([feature.segment_ids for feature in features]).reshape(-1, max_seq_length)

                    pred = sess.run(tensor_outputs, feed_dict={
                            tensor_input_ids: input_ids,
                            tensor_input_mask: input_mask,
                            tensor_label_ids: label_ids,
                            tensor_segment_ids: segment_ids,
                        })
                    predictions.extend(pred)

        return predictions
    
    
    def text_inference(self, texts):

        max_seq_length, tokenizer = self.Configs.max_seq_length, self.tokenizer

        graph = tf.Graph()
        with graph.as_default():
            with tf.Session() as sess:

                tf.saved_model.loader.load(sess, [tag_constants.SERVING], self.Configs.export_dir)
                tensor_input_ids = graph.get_tensor_by_name('input_ids_1:0')
                tensor_input_mask = graph.get_tensor_by_name('input_mask_1:0')
                tensor_label_ids = graph.get_tensor_by_name('label_ids_1:0')
                tensor_segment_ids = graph.get_tensor_by_name('segment_ids_1:0')
                tensor_outputs = graph.get_tensor_by_name('Sigmoid:0')

                
                features = [self.convert_test_single_example(txt, max_seq_length, tokenizer) for txt in texts]
                input_ids = np.array([feature.input_ids for feature in features]).reshape(-1, max_seq_length)
                input_mask = np.array([feature.input_mask for feature in features]).reshape(-1, max_seq_length)
                label_ids = np.array([feature.label_id for feature in features])
                segment_ids = np.array([feature.segment_ids for feature in features]).reshape(-1, max_seq_length)

                pred = sess.run(tensor_outputs, feed_dict={
                        tensor_input_ids: input_ids,
                        tensor_input_mask: input_mask,
                        tensor_label_ids: label_ids,
                        tensor_segment_ids: segment_ids,
                    })
                
        return pred


# coding=utf-8
# Copyright 2020 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions and classes related to optimization (weight updates).
Modified from the original BERT code to allow for having separate learning
rates for different layers of the network.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import re
import tensorflow.compat.v1 as tf


def create_optimizer(
    loss, learning_rate, num_train_steps, weight_decay_rate=0.0, use_tpu=False,
    warmup_steps=0, warmup_proportion=0, lr_decay_power=1.0,
    layerwise_lr_decay_power=-1, n_transformer_layers=None):
  """Creates an optimizer and training op."""
  global_step = tf.train.get_or_create_global_step()
  learning_rate = tf.train.polynomial_decay(
      learning_rate,
      global_step,
      num_train_steps,
      end_learning_rate=0.0,
      power=lr_decay_power,
      cycle=False)
  warmup_steps = max(num_train_steps * warmup_proportion, warmup_steps)
  learning_rate *= tf.minimum(
      1.0, tf.cast(global_step, tf.float32) / tf.cast(warmup_steps, tf.float32))

  if layerwise_lr_decay_power > 0:
    learning_rate = _get_layer_lrs(learning_rate, layerwise_lr_decay_power,
                                   n_transformer_layers)
  optimizer = AdamWeightDecayOptimizer(
      learning_rate=learning_rate,
      weight_decay_rate=weight_decay_rate,
      beta_1=0.9,
      beta_2=0.999,
      epsilon=1e-6,
      exclude_from_weight_decay=["LayerNorm", "layer_norm", "bias"])
  if use_tpu:
    optimizer = tf.tpu.CrossShardOptimizer(optimizer)

  tvars = tf.trainable_variables()
  grads = tf.gradients(loss, tvars)
  (grads, _) = tf.clip_by_global_norm(grads, clip_norm=1.0)
  train_op = optimizer.apply_gradients(
      zip(grads, tvars), global_step=global_step)
  new_global_step = global_step + 1
  train_op = tf.group(train_op, [global_step.assign(new_global_step)])
  return train_op


class AdamWeightDecayOptimizer(tf.train.Optimizer):
  """A basic Adam optimizer that includes "correct" L2 weight decay."""

  def __init__(self,
               learning_rate,
               weight_decay_rate=0.0,
               beta_1=0.9,
               beta_2=0.999,
               epsilon=1e-6,
               exclude_from_weight_decay=None,
               name="AdamWeightDecayOptimizer"):
    """Constructs a AdamWeightDecayOptimizer."""
    super(AdamWeightDecayOptimizer, self).__init__(False, name)

    self.learning_rate = learning_rate
    self.weight_decay_rate = weight_decay_rate
    self.beta_1 = beta_1
    self.beta_2 = beta_2
    self.epsilon = epsilon
    self.exclude_from_weight_decay = exclude_from_weight_decay

  def _apply_gradients(self, grads_and_vars, learning_rate):
    """See base class."""
    assignments = []
    for (grad, param) in grads_and_vars:
      if grad is None or param is None:
        continue

      param_name = self._get_variable_name(param.name)

      m = tf.get_variable(
          name=param_name + "/adam_m",
          shape=param.shape.as_list(),
          dtype=tf.float32,
          trainable=False,
          initializer=tf.zeros_initializer())
      v = tf.get_variable(
          name=param_name + "/adam_v",
          shape=param.shape.as_list(),
          dtype=tf.float32,
          trainable=False,
          initializer=tf.zeros_initializer())

      # Standard Adam update.
      next_m = (
          tf.multiply(self.beta_1, m) + tf.multiply(1.0 - self.beta_1, grad))
      next_v = (
          tf.multiply(self.beta_2, v) + tf.multiply(1.0 - self.beta_2,
                                                    tf.square(grad)))
      update = next_m / (tf.sqrt(next_v) + self.epsilon)

      # Just adding the square of the weights to the loss function is *not*
      # the correct way of using L2 regularization/weight decay with Adam,
      # since that will interact with the m and v parameters in strange ways.
      #
      # Instead we want ot decay the weights in a manner that doesn't interact
      # with the m/v parameters. This is equivalent to adding the square
      # of the weights to the loss with plain (non-momentum) SGD.
      if self.weight_decay_rate > 0:
        if self._do_use_weight_decay(param_name):
          update += self.weight_decay_rate * param

      update_with_lr = learning_rate * update
      next_param = param - update_with_lr

      assignments.extend(
          [param.assign(next_param),
           m.assign(next_m),
           v.assign(next_v)])

    return assignments

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    if isinstance(self.learning_rate, dict):
      key_to_grads_and_vars = {}
      for grad, var in grads_and_vars:
        update_for_var = False
        for key in self.learning_rate:
          if key in var.name:
            update_for_var = True
            if key not in key_to_grads_and_vars:
              key_to_grads_and_vars[key] = []
            key_to_grads_and_vars[key].append((grad, var))
        if not update_for_var:
          raise ValueError("No learning rate specified for variable", var)
      assignments = []
      for key, key_grads_and_vars in key_to_grads_and_vars.items():
        assignments += self._apply_gradients(key_grads_and_vars,
                                             self.learning_rate[key])
    else:
      assignments = self._apply_gradients(grads_and_vars, self.learning_rate)
    return tf.group(*assignments, name=name)

  def _do_use_weight_decay(self, param_name):
    """Whether to use L2 weight decay for `param_name`."""
    if not self.weight_decay_rate:
      return False
    if self.exclude_from_weight_decay:
      for r in self.exclude_from_weight_decay:
        if re.search(r, param_name) is not None:
          return False
    return True

  def _get_variable_name(self, param_name):
    """Get the variable name from the tensor name."""
    m = re.match("^(.*):\\d+$", param_name)
    if m is not None:
      param_name = m.group(1)
    return param_name


def _get_layer_lrs(learning_rate, layer_decay, n_layers):
  """Have lower learning rates for layers closer to the input."""
  key_to_depths = collections.OrderedDict({
      "/embeddings/": 0,
      "/embeddings_project/": 0,
      "task_specific/": n_layers + 2,
  })
  for layer in range(n_layers):
    key_to_depths["encoder/layer_" + str(layer) + "/"] = layer + 1
  return {
      key: learning_rate * (layer_decay ** (n_layers + 2 - depth))
      for key, depth in key_to_depths.items()
  }



# Copyright 2020 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifer: Apache-2.0

"""Set configurations for language models on a downstream classification task."""

import os

class Configs(object):
    
    """Fine-tuning hyperparameters."""


    """
    Parameters
    ----------
    
    pretrained_model_dir:   Directory path of pretrained models. (check downloader.py)
    
    model_name:             Name of the language model in small case (as string)
    
    learning_rate:          The initial learning rate for Adam.
    
    num_train_epochs:       Total number of training epochs to perform.
    
    train_batch_size:       Total batch size for training.
    
    eval_batch_size:        Total batch size for eval.
    
    predict_batch_size:     Total batch size for predict.
    
    do_train:               Whether to run training.
    
    do_eval:                Whether to run eval on the dev set.
    
    do_predict:             Whether to run training.
    
    label_list:             Number of classes in data
    
    do_lower_case:          Whether to lower case the input text. 
                            Should be True for uncased models and False for cased models
    
    max_seq_length:         The maximum total input sequence length after WordPiece tokenization. 
                            Sequences longer than this will be truncated, and shorter than this will be padded.
    
    data_dir:               The input data dir. Should contain the .csv files 
                            (train.csv, dev.csv, test.csv) for the task.
    
    output_dir:             The output directory where the model checkpoints will be written.
    
    num_tpu_cores:          Only used if `use_tpu` is True. Total number of TPU cores to use.
    
    gcp_project:            [Optional] Project name for the Cloud TPU-enabled project. 
                            If not specified, we will attempt to automatically detect the GCE project from metadata.
    
    master:                 [Optional] TensorFlow master URL.
    
    tpu_zone:               [Optional] GCE zone where the Cloud TPU is located in. 
                            If not specified, we will attempt to automatically detect the GCE project from metadata.
    
    tpu_name:               [Optional] GCE zone where the Cloud TPU is located in. 
                            If not specified, we will attempt to automatically detect the GCE project from metadata.
    
    use_tpu:                Whether to use TPU or GPU/CPU.
    
    iterations_per_loop:    How many steps to make in each estimator call.
    
    save_checkpoints_steps: How often to save the model checkpoint. 
    
    warmup_proportion:      Proportion of training to perform linear learning rate warmup for. 
                            E.g., 0.1 = 10% of training.
    
    export_dir:             The output model (.pb format) dir, where the freezed graph and weights get stored (only for BERT).
    
    tokenizer:              Wordpiece tokenizer object.
    
    export_path:            Path of saved .pb model for inferencing purpose (Only for BERT).
    
    """


    def __init__(self, pretrained_model_dir = './Albert',
              data_dir = "./data/",
              output_dir = "./albert_output/",
              export_dir = None,
              model_name="", model_size ="base",
              learning_rate = 5e-5, num_train_epochs=3.0, train_batch_size = 16,
              eval_batch_size = 8, predict_batch_size = 8, do_train = True,
              do_eval = True, do_predict = False, label_list = ["0", "1"],
              do_lower_case = True, max_seq_length = 256, use_tpu = False,
              iterations_per_loop = 1000, save_checkpoint_steps = 1000000,
              warmup_proportion = 0.1, export_path ='./exported_bert_model'):
        
        
        # default locations of required files
        self.data_dir = data_dir
        self.output_dir = output_dir
        self.model_name = model_name
        self.pretrained_model_dir = pretrained_model_dir
        
        # general
        if self.model_name == "":
            raise ValueError('Kindly specify the model name (bert, electra, albert, roberta, distilbert)')
        
        if not os.path.exists(self.data_dir):
                raise ValueError(self.data_dir, ' not found !!')
        
        if not os.path.exists(self.pretrained_model_dir):
            raise ValueError(pretrained_model_dir, ' pretrained model not found !!')
            
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
                
        # Common paramteres for all models
        
        
        self.do_train = do_train  # train a model ?
        self.do_eval = do_eval  # evaluate the model ?
        self.do_predict = do_predict
        self.label_list = label_list
        
        self.num_train_epochs = num_train_epochs  # passes over the dataset during training
        self.warmup_proportion = warmup_proportion  # how much of training to warm up the LR for
        self.learning_rate = learning_rate
        self.do_lower_case = do_lower_case
        
        # Params related to sizes, common to all
        
        self.max_seq_length = max_seq_length
        self.train_batch_size = train_batch_size
        self.eval_batch_size = eval_batch_size
        self.predict_batch_size = predict_batch_size
        
        # model
        self.embedding_size = None  # bert hidden size by default
        self.vocab_size = 30522  # number of tokens in the vocabulary

        # training
        self.weight_decay_rate = 0.01
        self.layerwise_lr_decay = 0.8      
        self.save_checkpoints_steps = save_checkpoint_steps
        self.iterations_per_loop = iterations_per_loop
        self.use_tfrecords_if_existing = False    

        # TPU settings: Not required, just default init (required for execution)
        self.use_tpu = False
        self.num_tpu_cores = 1
        self.tpu_job_name = None
        self.tpu_name = None  # cloud TPU to use for training
        self.tpu_zone = None  # GCE zone where the Cloud TPU is located in
        self.gcp_project = None  # project name for the Cloud TPU-enabled project
        self.master = None

        self.vocab_file = os.path.join(pretrained_model_dir, "vocab.txt")
        
        # Specific to BERT tensorflow model
        if self.model_name == 'bert':
            self.init_checkpoint = os.path.join(pretrained_model_dir,'bert_model.ckpt')
            self.bert_config_file = os.path.join(pretrained_model_dir,'bert_config.json')
            if not os.path.exists(self.bert_config_file):
                raise ValueError('bert_config.json not found in pretrained_model_dir!!')
            if not os.path.exists(self.vocab_file):
                raise ValueError('vocab.txt not found in pretrained_model_dir!!')

            self.export_path = export_path
            self.export_dir = export_dir

            if not os.path.exists(self.export_path):
                os.makedirs(self.export_path)
            
        # Specific to Electra Tensorflow model
        elif self.model_name == 'electra':
            self.init_checkpoint = pretrained_model_dir
            self.model_size = model_size  # one of "small", "base", or "large"
            self.task_names = ["textclassification"]  # which tasks to learn
            self.model_hparam_overrides = {}
            self.num_trials = 1  # how many train+eval runs to perform
            self.keep_all_models = True
            self.log_examples = False

            if not os.path.exists(self.vocab_file):
                raise ValueError('vocab.txt not found in pretrained_model_dir!!')

            # default hyperparameters for different model sizes
            if self.model_size == "large":
                self.learning_rate = 5e-5
                self.layerwise_lr_decay = 0.9
            elif self.model_size == "small":
                self.embedding_size = 128

            # update defaults with passed-in hyperparameters
            self.tasks = {
              "textclassification":{
                "type":"classification",
                "labels":label_list,
                "header":True,
                "text_column":1,
                "label_column":2
                }
            }



# Copyright 2020 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifer: Apache-2.0

import os
import wget
import shutil
from zipfile import ZipFile


def downloader(model, destination_path='../models'):
    
    """
    Function to download pre-trained models from hugginf face aws repository OR Google's storgage.

    Parameters
    ----------
    model : Pick one of these ['bert','electra','roberta','distilbert','albert']
    download_path : Directory whether the model has to be downloaded

    Returns
    -------

    """
    
    model_download = {
            "bert": 'Bert_base',
            "electra": 'Electra_base',
            "roberta": 'Roberta',
            "distilbert" : 'DistilBert',
            "albert": 'Albert'
    }
    
    if model not in model_download:
        print("Please pick model name from ['bert','electra','roberta','distilbert','albert']")
        return None

    output_dir = os.path.join(destination_path, model_download[model])
    print("Model gets downloaded here: ", output_dir)

    # Create output directory if needed
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
        
    if model == 'roberta':
        
        
        config_file = wget.download('https://s3.amazonaws.com/models.huggingface.co/bert/roberta-base-config.json',
                                    os.path.join(output_dir,'config.json'))
        vocab_file = wget.download('https://s3.amazonaws.com/models.huggingface.co/bert/roberta-base-vocab.json',
                                    os.path.join(output_dir,'vocab.json'))
        merges_file = wget.download('https://s3.amazonaws.com/models.huggingface.co/bert/roberta-base-merges.txt',
                                    os.path.join(output_dir,'merges.txt'))
        model_file = wget.download('https://s3.amazonaws.com/models.huggingface.co/bert/roberta-base-pytorch_model.bin',
                                   os.path.join(output_dir,'pytorch_model.bin'))
    
    if model == 'albert':
        config_file = wget.download('https://s3.amazonaws.com/models.huggingface.co/bert/albert-base-v2-config.json',
                                    os.path.join(output_dir,'config.json'))
        spiece_file = wget.download('https://s3.amazonaws.com/models.huggingface.co/bert/albert-base-v2-spiece.model',
                                    os.path.join(output_dir,'spiece.model'))
        model_file = wget.download('https://s3.amazonaws.com/models.huggingface.co/bert/albert-base-v2-pytorch_model.bin',
                                   os.path.join(output_dir,'pytorch_model.bin'))
        
    if model == 'distilbert':
        config_file = wget.download('https://s3.amazonaws.com/models.huggingface.co/bert/distilbert-base-uncased-config.json',
                                    os.path.join(output_dir,'config.json'))
        vocab_file = wget.download('https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-uncased-vocab.txt',
                                    os.path.join(output_dir,'vocab.json'))
        model_file = wget.download('https://s3.amazonaws.com/models.huggingface.co/bert/distilbert-base-uncased-pytorch_model.bin',
                                   os.path.join(output_dir,'pytorch_model.bin'))
    
    
    if model == 'bert':
        zip_file = wget.download('https://storage.googleapis.com/bert_models/2020_02_20/uncased_L-12_H-768_A-12.zip',
                                    os.path.join(output_dir,'uncased_L-12_H-768_A-12.zip'))
        
        with ZipFile(os.path.join(output_dir,'uncased_L-12_H-768_A-12.zip'), "r") as zip_ref:
            zip_ref.extractall(output_dir)

        os.remove(os.path.join(output_dir,'uncased_L-12_H-768_A-12.zip'))
        
    if model == 'electra':
        zip_file = wget.download('https://storage.googleapis.com/electra-data/electra_base.zip',
                                  os.path.join(output_dir,'electra_base.zip'))
        
        with ZipFile(os.path.join(output_dir,'electra_base.zip'), "r") as zip_ref:
            zip_ref.extractall(output_dir)
            
        files = os.listdir(os.path.join(output_dir,'electra_base'))
        for f in files:
            shutil.move(os.path.join(output_dir,'electra_base',f), output_dir)

        os.remove(os.path.join(output_dir,'electra_base.zip'))
        os.rmdir(os.path.join(output_dir,'electra_base'))


# Copyright 2020 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifer: Apache-2.0

"""Fine-tunes an ELECTRA model on a downstream task."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import os
import abc
import csv
import sys
import json
import time
import scipy
import random
import sklearn
import datetime
import collections
import numpy as np
import pandas as pd
from typing import List, Tuple
import tensorflow.compat.v1 as tf
from classitransformers.configs import Configs


from electra import modeling
from electra import optimization
from electra import tokenization



def get_shared_feature_specs(config: Configs):
  """Non-task-specific model inputs."""
  return [
      FeatureSpec("input_ids", [config.max_seq_length]),
      FeatureSpec("input_mask", [config.max_seq_length]),
      FeatureSpec("segment_ids", [config.max_seq_length]),
      FeatureSpec("task_id", []),
  ]


class FeatureSpec(object):
  """Defines a feature passed as input to the model."""

  def __init__(self, name, shape, default_value_fn=None, is_int_feature=True):
    self.name = name
    self.shape = shape
    self.default_value_fn = default_value_fn
    self.is_int_feature = is_int_feature

  def get_parsing_spec(self):
    return tf.io.FixedLenFeature(
        self.shape, tf.int64 if self.is_int_feature else tf.float32)

  def get_default_values(self):
    if self.default_value_fn:
      return self.default_value_fn(self.shape)
    else:
      return np.zeros(
          self.shape, np.int64 if self.is_int_feature else np.float32)



class Preprocessor(object):
  """Class for loading, preprocessing, and serializing fine-tuning datasets."""

  def __init__(self, config: Configs, tasks):
    self._config = config
    self._tasks = tasks
    self._name_to_task = {task.name: task for task in tasks}

    self._feature_specs = get_shared_feature_specs(config)
    for task in tasks:
      self._feature_specs += task.get_feature_specs()
    self._name_to_feature_config = {
        spec.name: spec.get_parsing_spec()
        for spec in self._feature_specs
    }
    assert len(self._name_to_feature_config) == len(self._feature_specs)

  def prepare_train(self):
    return self._serialize_dataset(self._tasks, True, "train")

  def prepare_predict(self, tasks, split):
    return self._serialize_dataset(tasks, False, split)

  def _serialize_dataset(self, tasks, is_training, split):
    """Write out the dataset as tfrecords."""
    dataset_name = "_".join(sorted([task.name for task in tasks]))
    dataset_name += "_" + split
    dataset_prefix = os.path.join(
        self._config.output_dir, dataset_name)
    tfrecords_path = dataset_prefix + ".tfrecord"
    metadata_path = dataset_prefix + ".metadata"
    batch_size = (self._config.train_batch_size if is_training else
                  self._config.eval_batch_size)

    log("Loading dataset", dataset_name)
    n_examples = None
    if (self._config.use_tfrecords_if_existing and
        tf.io.gfile.exists(metadata_path)) and split != 'infer':
      n_examples = load_json(metadata_path)["n_examples"]

    if n_examples is None:
      log("Existing tfrecords not found so creating")
      examples = []
      for task in tasks:
        task_examples = task.get_examples(split)
        examples += task_examples
      if is_training:
        random.shuffle(examples)
      mkdir(tfrecords_path.rsplit("/", 1)[0])
      n_examples = self.serialize_examples(
          examples, is_training, tfrecords_path, batch_size)
      write_json({"n_examples": n_examples}, metadata_path)

    input_fn = self._input_fn_builder(tfrecords_path, is_training)
    if is_training:
      steps = int(n_examples // batch_size * self._config.num_train_epochs)
    else:
      steps = n_examples // batch_size

    return input_fn, steps

  def serialize_examples(self, examples, is_training, output_file, batch_size):
    """Convert a set of `InputExample`s to a TFRecord file."""
    n_examples = 0
    with tf.io.TFRecordWriter(output_file) as writer:
      for (ex_index, example) in enumerate(examples):
        if ex_index % 2000 == 0:
          log("Writing example {:} of {:}".format(
              ex_index, len(examples)))
        for tf_example in self._example_to_tf_example(
            example, is_training,
            log=self._config.log_examples and ex_index < 1):
          writer.write(tf_example.SerializeToString())
          n_examples += 1
      # add padding so the dataset is a multiple of batch_size
      while n_examples % batch_size != 0:
        writer.write(self._make_tf_example(task_id=len(self._config.task_names))
                     .SerializeToString())
        n_examples += 1
    return n_examples

  def _example_to_tf_example(self, example, is_training, log=False):
    examples = self._name_to_task[example.task_name].featurize(
        example, is_training, log)
    if not isinstance(examples, list):
      examples = [examples]
    for example in examples:
      yield self._make_tf_example(**example)

  def _make_tf_example(self, **kwargs):
    """Make a tf.train.Example from the provided features."""
    for k in kwargs:
      if k not in self._name_to_feature_config:
        raise ValueError("Unknown feature", k)
    features = collections.OrderedDict()
    for spec in self._feature_specs:
      if spec.name in kwargs:
        values = kwargs[spec.name]
      else:
        values = spec.get_default_values()
      if (isinstance(values, int) or isinstance(values, bool) or
          isinstance(values, float) or isinstance(values, np.float32) or
          (isinstance(values, np.ndarray) and values.size == 1)):
        values = [values]
      if spec.is_int_feature:
        feature = tf.train.Feature(int64_list=tf.train.Int64List(
            value=list(values)))
      else:
        feature = tf.train.Feature(float_list=tf.train.FloatList(
            value=list(values)))
      features[spec.name] = feature
    return tf.train.Example(features=tf.train.Features(feature=features))

  def _input_fn_builder(self, input_file, is_training):
    """Creates an `input_fn` closure to be passed to TPUEstimator."""

    def input_fn(params):
      """The actual input function."""
      d = tf.data.TFRecordDataset(input_file)
      if is_training:
        d = d.repeat()
        d = d.shuffle(buffer_size=100)
      return d.apply(
          tf.data.experimental.map_and_batch(
              self._decode_tfrecord,
              batch_size=params["batch_size"],
              drop_remainder=True))

    return input_fn

  def _decode_tfrecord(self, record):
    """Decodes a record to a TensorFlow example."""
    example = tf.io.parse_single_example(record, self._name_to_feature_config)
    # tf.Example only supports tf.int64, but the TPU only supports tf.int32.
    # So cast all int64 to int32.
    for name, tensor in example.items():
      if tensor.dtype == tf.int64:
        example[name] = tf.cast(tensor, tf.int32)
      else:
        example[name] = tensor
    return example

class Scorer(object):
  """Abstract base class for computing evaluation metrics."""

  __metaclass__ = abc.ABCMeta

  def __init__(self):
    self._updated = False
    self._cached_results = {}

  @abc.abstractmethod
  def update(self, results):
    self._updated = True

  @abc.abstractmethod
  def get_loss(self):
    pass

  @abc.abstractmethod
  def _get_results(self):
    return []

  def get_results(self, prefix=""):
    results = self._get_results() if self._updated else self._cached_results
    self._cached_results = results
    self._updated = False
    return [(prefix + k, v) for k, v in results]

  def results_str(self):
    return " - ".join(["{:}: {:.2f}".format(k, v)
                       for k, v in self.get_results()])


class SentenceLevelScorer(Scorer):
  """Abstract scorer for classification/regression tasks."""

  __metaclass__ = abc.ABCMeta

  def __init__(self):
    super(SentenceLevelScorer, self).__init__()
    self._total_loss = 0
    self._true_labels = []
    self._preds = []

  def update(self, results):
    super(SentenceLevelScorer, self).update(results)
    self._total_loss += results['loss']
    self._true_labels.append(results['label_ids'] if 'label_ids' in results
                             else results['targets'])
    self._preds.append(results['predictions'])

  def get_loss(self):
    return self._total_loss / len(self._true_labels)


class AccuracyScorer(SentenceLevelScorer):

  def _get_results(self):
    correct, count = 0, 0
    for y_true, pred in zip(self._true_labels, self._preds):
      count += 1
      correct += (1 if y_true == pred else 0)
    return [
        ('accuracy', 100.0 * correct / count),
        ('loss', self.get_loss()),
    ]



class Example(object):
  __metaclass__ = abc.ABCMeta

  def __init__(self, task_name):
    self.task_name = task_name


class Task(object):
  """Override this class to add a new fine-tuning task."""

  __metaclass__ = abc.ABCMeta

  def __init__(self, config: Configs, name):
    self.config = config
    self.name = name

  def get_test_splits(self):
    return ["test"]

  @abc.abstractmethod
  def get_examples(self, split):
    pass

  @abc.abstractmethod
  def get_scorer(self) -> Scorer:
    pass

  @abc.abstractmethod
  def get_feature_specs(self) -> List[FeatureSpec]:
    pass

  @abc.abstractmethod
  def featurize(self, example: Example, is_training: bool,
                log: bool=False):
    pass

  @abc.abstractmethod
  def get_prediction_module(
      self, bert_model: modeling.BertModel, features: dict, is_training: bool,
      percent_done: float) -> Tuple:
    pass

  def __repr__(self):
    return "Task(" + self.name + ")"




class InputExample(Example):
  """A single training/test example for simple sequence classification."""

  def __init__(self, eid, task_name, text_a, text_b=None, label=None):
    super(InputExample, self).__init__(task_name)
    self.eid = eid
    self.text_a = text_a
    self.text_b = text_b
    self.label = label


class SingleOutputTask(Task):
  """Task with a single prediction per example (e.g., text classification)."""

  __metaclass__ = abc.ABCMeta

  def __init__(self, config: Configs, name,
               tokenizer):
    super(SingleOutputTask, self).__init__(config, name)
    self._tokenizer = tokenizer

  def get_examples(self, split):
    return self._create_examples(read_csv(
        os.path.join(self.config.data_dir, split + ".csv"),
        max_lines= None), split)

  @abc.abstractmethod
  def _create_examples(self, lines, split):
    pass

  def featurize(self, example: InputExample, is_training, log=False):
    """Turn an InputExample into a dict of features."""
    tokens_a = self._tokenizer.tokenize(example.text_a)
    tokens_b = None
    if example.text_b:
      tokens_b = self._tokenizer.tokenize(example.text_b)

    if tokens_b:
      # Modifies `tokens_a` and `tokens_b` in place so that the total
      # length is less than the specified length.
      # Account for [CLS], [SEP], [SEP] with "- 3"
      _truncate_seq_pair(tokens_a, tokens_b, self.config.max_seq_length - 3)
    else:
      # Account for [CLS] and [SEP] with "- 2"
      if len(tokens_a) > self.config.max_seq_length - 2:
        tokens_a = tokens_a[0:(self.config.max_seq_length - 2)]

    # The convention in BERT is:
    # (a) For sequence pairs:
    #  tokens:   [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
    #  type_ids: 0     0  0    0    0     0       0 0     1  1  1  1   1 1
    # (b) For single sequences:
    #  tokens:   [CLS] the dog is hairy . [SEP]
    #  type_ids: 0     0   0   0  0     0 0
    #
    # Where "type_ids" are used to indicate whether this is the first
    # sequence or the second sequence. The embedding vectors for `type=0` and
    # `type=1` were learned during pre-training and are added to the wordpiece
    # embedding vector (and position vector). This is not *strictly* necessary
    # since the [SEP] token unambiguously separates the sequences, but it
    # makes it easier for the model to learn the concept of sequences.
    #
    # For classification tasks, the first vector (corresponding to [CLS]) is
    # used as the "sentence vector". Note that this only makes sense because
    # the entire model is fine-tuned.
    tokens = []
    segment_ids = []
    tokens.append("[CLS]")
    segment_ids.append(0)
    for token in tokens_a:
      tokens.append(token)
      segment_ids.append(0)
    tokens.append("[SEP]")
    segment_ids.append(0)

    if tokens_b:
      for token in tokens_b:
        tokens.append(token)
        segment_ids.append(1)
      tokens.append("[SEP]")
      segment_ids.append(1)

    input_ids = self._tokenizer.convert_tokens_to_ids(tokens)

    # The mask has 1 for real tokens and 0 for padding tokens. Only real
    # tokens are attended to.
    input_mask = [1] * len(input_ids)

    # Zero-pad up to the sequence length.
    while len(input_ids) < self.config.max_seq_length:
      input_ids.append(0)
      input_mask.append(0)
      segment_ids.append(0)

    assert len(input_ids) == self.config.max_seq_length
    assert len(input_mask) == self.config.max_seq_length
    assert len(segment_ids) == self.config.max_seq_length

    if log:
      log("  Example {:}".format(example.eid))
      log("    tokens: {:}".format(" ".join(
          [tokenization.printable_text(x) for x in tokens])))
      log("    input_ids: {:}".format(" ".join(map(str, input_ids))))
      log("    input_mask: {:}".format(" ".join(map(str, input_mask))))
      log("    segment_ids: {:}".format(" ".join(map(str, segment_ids))))

    eid = example.eid
    features = {
        "input_ids": input_ids,
        "input_mask": input_mask,
        "segment_ids": segment_ids,
        "task_id": self.config.task_names.index(self.name),
        self.name + "_eid": eid,
    }
    self._add_features(features, example, log)
    return features

  def _load_glue(self, lines, split, text_a_loc, text_b_loc, label_loc,
                 skip_first_line=False, eid_offset=0, swap=False):
    examples = []
    for (i, line) in enumerate(lines):
      try:
        if i == 0 and skip_first_line:
          continue
        eid = i - (1 if skip_first_line else 0) + eid_offset
        text_a = tokenization.convert_to_unicode(line[text_a_loc])
        if text_b_loc is None:
          text_b = None
        else:
          text_b = tokenization.convert_to_unicode(line[text_b_loc])
        if "test" in split or "diagnostic" in split or "infer" in split:
          label = self._get_dummy_label()
        else:
          label = tokenization.convert_to_unicode(line[label_loc])
        if swap:
          text_a, text_b = text_b, text_a
        examples.append(InputExample(eid=eid, task_name=self.name,
                                     text_a=text_a, text_b=text_b, label=label))
      except Exception as ex:
        log("Error constructing example from line", i,
                  "for task", self.name + ":", ex)
        log("Input causing the error:", line)
    return examples

  @abc.abstractmethod
  def _get_dummy_label(self):
    pass

  @abc.abstractmethod
  def _add_features(self, features, example, log):
    pass



class ClassificationTask(SingleOutputTask):
  """Task where the output is a single categorical label for the input text."""
  __metaclass__ = abc.ABCMeta

  def __init__(self, config: Configs, name,
               tokenizer, label_list):
    super(ClassificationTask, self).__init__(config, name, tokenizer)
    self._tokenizer = tokenizer
    self._label_list = label_list

  def _get_dummy_label(self):
    return self._label_list[0]

  def get_feature_specs(self):
    return [FeatureSpec(self.name + "_eid", []),
            FeatureSpec(self.name + "_label_ids", [])]

  def _add_features(self, features, example, log):
    label_map = {}
    for (i, label) in enumerate(self._label_list):
      label_map[label] = i
    label_id = label_map[example.label]
    if log:
      log("    label: {:} (id = {:})".format(example.label, label_id))
    features[example.task_name + "_label_ids"] = label_id

  def get_prediction_module(self, bert_model, features, is_training,
                            percent_done):
    num_labels = len(self._label_list)
    reprs = bert_model.get_pooled_output()

    if is_training:
      reprs = tf.nn.dropout(reprs, keep_prob=0.9)

    logits = tf.layers.dense(reprs, num_labels)
    log_probs = tf.nn.log_softmax(logits, axis=-1)

    label_ids = features[self.name + "_label_ids"]
    labels = tf.one_hot(label_ids, depth=num_labels, dtype=tf.float32)

    losses = -tf.reduce_sum(labels * log_probs, axis=-1)

    outputs = dict(
        loss=losses,
        logits=logits,
        predictions=tf.argmax(logits, axis=-1),
        label_ids=label_ids,
        eid=features[self.name + "_eid"],
    )
    return losses, outputs

  def get_scorer(self):
    return AccuracyScorer()


def _truncate_seq_pair(tokens_a, tokens_b, max_length):
  """Truncates a sequence pair in place to the maximum length."""

  # This is a simple heuristic which will always truncate the longer sequence
  # one token at a time. This makes more sense than truncating an equal percent
  # of tokens from each, since if one sequence is very short then each token
  # that's truncated likely contains more information than a longer sequence.
  while True:
    total_length = len(tokens_a) + len(tokens_b)
    if total_length <= max_length:
      break
    if len(tokens_a) > len(tokens_b):
      tokens_a.pop()
    else:
      tokens_b.pop()


def read_csv(input_file, quotechar=None, max_lines=None):
  """Reads a tab separated value file."""
  with tf.io.gfile.GFile(input_file, "r") as f:
    reader = csv.reader(f, delimiter=",", quotechar=quotechar)
    lines = []
    for i, line in enumerate(reader):
      if max_lines and i >= max_lines:
        break
      lines.append(line)
    return lines


  
class StandardTSV(ClassificationTask):
  def __init__(self, config: Configs,
               task_name: str, task_config: dict, tokenizer):
    self.task_config = task_config
    labels = len(self.task_config.get("labels", None))
    labels_list = []
    for i in range(0, labels):
      labels_list.append(str(i))
    super(StandardTSV, self).__init__(config, task_name, tokenizer,
                               labels_list)
    #self.task_config = task_config

  def get_examples(self, split):
    return self._create_examples(read_csv(
        os.path.join(self.config.data_dir, split + ".csv"),
        quotechar = "\"",
        max_lines= None), split)

  def _create_examples(self, lines, split):
    text_column_2 = self.task_config.get("text_column_2", None)
    header = self.task_config.get("header", False)
    return self._load_glue(lines, split, self.task_config["text_column"],
                           text_column_2, self.task_config["label_column"],
                           skip_first_line=header)


def get_tasks(config: Configs):
  tokenizer = tokenization.FullTokenizer(vocab_file=config.vocab_file,
                                         do_lower_case=config.do_lower_case)
  return [get_task(config, task_name, tokenizer)
          for task_name in config.task_names]


def get_task(config: Configs, task_name,
             tokenizer):
  """Get an instance of a task based on its name."""
  if (task_name in config.tasks):
    if config.tasks[task_name]["type"] == "classification":
      return StandardTSV(config, task_name, config.tasks[task_name], tokenizer)
    else:
      raise ValueError("Unknown task type: " + config.tasks[task_name]["type"])
  else:
    raise ValueError("Unknown task " + task_name)



class FinetuningModel(object):
  """Finetuning model with support for multi-task training."""

  def __init__(self, config: Configs, tasks,
               is_training, features, num_train_steps):
    # Create a shared transformer encoder
    bert_config = get_bert_config(config)
    self.bert_config = bert_config

    assert config.max_seq_length <= bert_config.max_position_embeddings
    bert_model = modeling.BertModel(
        bert_config=bert_config,
        is_training=is_training,
        input_ids=features["input_ids"],
        input_mask=features["input_mask"],
        token_type_ids=features["segment_ids"],
        use_one_hot_embeddings=config.use_tpu,
        embedding_size=config.embedding_size)
    percent_done = (tf.cast(tf.train.get_or_create_global_step(), tf.float32) /
                    tf.cast(num_train_steps, tf.float32))

    # Add specific tasks
    self.outputs = {"task_id": features["task_id"]}
    losses = []
    for task in tasks:
      with tf.variable_scope("task_specific/" + task.name):
        task_losses, task_outputs = task.get_prediction_module(
            bert_model, features, is_training, percent_done)
        losses.append(task_losses)
        self.outputs[task.name] = task_outputs
    self.loss = tf.reduce_sum(
        tf.stack(losses, -1) *
        tf.one_hot(features["task_id"], len(config.task_names)))


def model_fn_builder(config: Configs, tasks,
                     num_train_steps, pretraining_config=None):
  """Returns `model_fn` closure for TPUEstimator."""

  def model_fn(features, labels, mode, params):
    """The `model_fn` for TPUEstimator."""
    log("Building model...")
    is_training = (mode == tf.estimator.ModeKeys.TRAIN)
    model = FinetuningModel(
        config, tasks, is_training, features, num_train_steps)

    # Load pre-trained weights from checkpoint
    init_checkpoint = config.init_checkpoint
    if pretraining_config is not None:
      init_checkpoint = tf.train.latest_checkpoint(pretraining_config.output_dir)
      log("Using checkpoint", init_checkpoint)
    tvars = tf.trainable_variables()
    scaffold_fn = None
    if init_checkpoint:
      assignment_map, _ = modeling.get_assignment_map_from_checkpoint(
          tvars, init_checkpoint)
      if config.use_tpu:
        def tpu_scaffold():
          tf.train.init_from_checkpoint(init_checkpoint, assignment_map)
          return tf.train.Scaffold()
        scaffold_fn = tpu_scaffold
      else:
        tf.train.init_from_checkpoint(init_checkpoint, assignment_map)

    # Build model for training or prediction
    if mode == tf.estimator.ModeKeys.TRAIN:
      train_op = optimization.create_optimizer(
          model.loss, config.learning_rate, num_train_steps,
          weight_decay_rate=config.weight_decay_rate,
          use_tpu=config.use_tpu,
          warmup_proportion=config.warmup_proportion,
          layerwise_lr_decay_power=config.layerwise_lr_decay,
          n_transformer_layers=model.bert_config.num_hidden_layers
      )
      output_spec = tf.estimator.tpu.TPUEstimatorSpec(
          mode=mode,
          loss=model.loss,
          train_op=train_op,
          scaffold_fn=scaffold_fn,
          training_hooks=[ETAHook(
              {} if config.use_tpu else dict(loss=model.loss),
              num_train_steps, config.iterations_per_loop, config.use_tpu, 10)])
    else:
      assert mode == tf.estimator.ModeKeys.PREDICT
      output_spec = tf.estimator.tpu.TPUEstimatorSpec(
          mode=mode,
          predictions=flatten_dict(model.outputs),
          scaffold_fn=scaffold_fn)

    log("Building complete")
    return output_spec

  return model_fn


class ElectraClassification(object):
  """Fine-tunes a model on a supervised task."""

  def __init__(self, config: Configs, pretraining_config=None):
    tf.logging.set_verbosity(tf.logging.ERROR)
    trial = 1
    heading_info = "model={:}, trial {:}/{:}".format(config.model_name, trial, config.num_trials)
    heading2 = lambda msg: heading(msg + ": " + heading_info)
    heading2("Config")
    log_config(config)
    generic_model_dir = config.output_dir
    tasks = get_tasks(config)
    
    config.output_dir = generic_model_dir 
    if config.do_train:
      rmkdir(config.output_dir)
      
    self._config = config
    self._tasks = tasks
    self._preprocessor = Preprocessor(config, self._tasks)

    is_per_host = tf.estimator.tpu.InputPipelineConfig.PER_HOST_V2
    tpu_cluster_resolver = None
    if config.use_tpu and config.tpu_name:
      tpu_cluster_resolver = tf.distribute.cluster_resolver.TPUClusterResolver(
          config.tpu_name, zone=config.tpu_zone, project=config.gcp_project)
    tpu_config = tf.estimator.tpu.TPUConfig(
        iterations_per_loop=config.iterations_per_loop,
        num_shards=config.num_tpu_cores,
        per_host_input_for_training=is_per_host,
        tpu_job_name=config.tpu_job_name)
    run_config = tf.estimator.tpu.RunConfig(
        cluster=tpu_cluster_resolver,
        model_dir=config.output_dir,
        save_checkpoints_steps=config.save_checkpoints_steps,
        save_checkpoints_secs=None,
        tpu_config=tpu_config)

    if self._config.do_train:
      (self._train_input_fn,
       self.train_steps) = self._preprocessor.prepare_train()
    else:
      self._train_input_fn, self.train_steps = None, 0
    model_fn = model_fn_builder(
        config=config,
        tasks=self._tasks,
        num_train_steps=self.train_steps,
        pretraining_config=pretraining_config)
    self._estimator = tf.estimator.tpu.TPUEstimator(
        use_tpu=config.use_tpu,
        model_fn=model_fn,
        config=run_config,
        train_batch_size=config.train_batch_size,
        eval_batch_size=config.eval_batch_size,
        predict_batch_size=config.predict_batch_size)

  def train(self):
    heading("Start training")
    log("Training for {:} steps".format(self.train_steps))
    self._estimator.train(input_fn=self._train_input_fn, max_steps=self.train_steps)
    heading("Training Finished")
    if self._config.do_eval:
      self.evaluate()
    

  def evaluate(self):
    heading("Start Evaluation")
    return {task.name: self.evaluate_task(task) for task in self._tasks}
  
  def test(self):
    tasks = get_tasks(self._config)
    for task in tasks:
        for split in task.get_test_splits():
            logits = self.write_classification_outputs([task], 1, split)
    
    preds = []
    for i in logits:
        preds.append(scipy.special.expit(i))
    return preds   

  def evaluate_task(self, task, split="dev", return_results=True):
    """Evaluate the current model."""
    log("Evaluating", task.name)
    eval_input_fn, _ = self._preprocessor.prepare_predict([task], split)
    results = self._estimator.predict(input_fn=eval_input_fn,
                                      yield_single_examples=True)
    scorer = task.get_scorer()
    for r in results:
      if r["task_id"] != len(self._tasks):  # ignore padding examples
        r = nest_dict(r, self._config.task_names)
        scorer.update(r[task.name])
    if return_results:
      log(task.name + ": " + scorer.results_str())
      log()
      return dict(scorer.get_results())
    else:
      return scorer


  def write_classification_outputs(self, tasks, trial, split):
    """Write classification predictions to disk."""
    log("Writing out predictions for", tasks, split)
    predict_input_fn, _ = self._preprocessor.prepare_predict(tasks, split)
    results = self._estimator.predict(input_fn=predict_input_fn,
                                      yield_single_examples=True)
    # task name -> eid -> model-logits
    logits = collections.defaultdict(dict)
    for r in results:
      if r["task_id"] != len(self._tasks):
        r = nest_dict(r, self._config.task_names)
        task_name = self._config.task_names[r["task_id"]]
        logits[task_name][r[task_name]["eid"]] = (
            r[task_name]["logits"] if "logits" in r[task_name]
            else r[task_name]["predictions"])
    for task_name in logits:
      log("Getting predictions for {:} {:} examples ({:})".format(
          len(logits[task_name]), task_name, split))
      return list(logits[task_name].values())
  

  def text_inference(self, texts):
    
    """ function to inference on list of sentences"""
    
    df = pd.DataFrame({"id": np.arange(len(texts)), "text":texts})
    df.to_csv(os.path.join(self._config.data_dir, "infer.csv"), index=False, sep=",")
    tasks = get_tasks(self._config)
    for task in tasks:
        logits = self.write_classification_outputs([task], 1, "infer")
    preds = []
    for i in logits:
        preds.append(scipy.special.expit(i))
        
    return preds


    
def write_results(config: Configs, results):
  """Write evaluation metrics to disk."""
  log("Writing results to", config.results_txt)
  mkdir(config.results_txt.rsplit("/", 1)[0])
  write_pickle(results, config.results_pkl)
  with tf.io.gfile.GFile(config.results_txt, "w") as f:
    results_str = ""
    for trial_results in results:
      for task_name, task_results in trial_results.items():
        if task_name == "time" or task_name == "global_step":
          continue
        results_str += task_name + ": " + " - ".join(
            ["{:}: {:.2f}".format(k, v)
             for k, v in task_results.items()]) + "\n"
    f.write(results_str)
  write_pickle(results, config.results_pkl)



def load_json(path):
  with tf.io.gfile.GFile(path, "r") as f:
    return json.load(f)


def write_json(o, path):
  if "/" in path:
    tf.io.gfile.makedirs(path.rsplit("/", 1)[0])
  with tf.io.gfile.GFile(path, "w") as f:
    json.dump(o, f)


def load_pickle(path):
  with tf.io.gfile.GFile(path, "rb") as f:
    return pickle.load(f)


def write_pickle(o, path):
  if "/" in path:
    tf.io.gfile.makedirs(path.rsplit("/", 1)[0])
  with tf.io.gfile.GFile(path, "wb") as f:
    pickle.dump(o, f, -1)


def mkdir(path):
  if not tf.io.gfile.exists(path):
    tf.io.gfile.makedirs(path)


def rmrf(path):
  if tf.io.gfile.exists(path):
    tf.io.gfile.rmtree(path)


def rmkdir(path):
  rmrf(path)
  mkdir(path)


def log(*args):
  msg = " ".join(map(str, args))
  sys.stdout.write(msg + "\n")
  sys.stdout.flush()


def log_config(config):
  for key, value in sorted(config.__dict__.items()):
    log(key, value)
  log()


def heading(*args):
  log(80 * "=")
  log(*args)
  log(80 * "=")


def nest_dict(d, prefixes, delim="_"):
  """Go from {prefix_key: value} to {prefix: {key: value}}."""
  nested = {}
  for k, v in d.items():
    for prefix in prefixes:
      if k.startswith(prefix + delim):
        if prefix not in nested:
          nested[prefix] = {}
        nested[prefix][k.split(delim, 1)[1]] = v
      else:
        nested[k] = v
  return nested


def flatten_dict(d, delim="_"):
  """Go from {prefix: {key: value}} to {prefix_key: value}."""
  flattened = {}
  for k, v in d.items():
    if isinstance(v, dict):
      for k2, v2 in v.items():
        flattened[k + delim + k2] = v2
    else:
      flattened[k] = v
  return flattened


class ETAHook(tf.estimator.SessionRunHook):
  """Print out the time remaining during training/evaluation."""

  def __init__(self, to_log, n_steps, iterations_per_loop, on_tpu,
               log_every=1, is_training=True):
    self._to_log = to_log
    self._n_steps = n_steps
    self._iterations_per_loop = iterations_per_loop
    self._on_tpu = on_tpu
    self._log_every = log_every
    self._is_training = is_training
    self._steps_run_so_far = 0
    self._global_step = None
    self._global_step_tensor = None
    self._start_step = None
    self._start_time = None

  def begin(self):
    self._global_step_tensor = tf.train.get_or_create_global_step()

  def before_run(self, run_context):
    if self._start_time is None:
      self._start_time = time.time()
    return tf.estimator.SessionRunArgs(self._to_log)

  def after_run(self, run_context, run_values):
    self._global_step = run_context.session.run(self._global_step_tensor)
    self._steps_run_so_far += self._iterations_per_loop if self._on_tpu else 1
    if self._start_step is None:
      self._start_step = self._global_step - (self._iterations_per_loop
                                              if self._on_tpu else 1)
    self.log(run_values)

  def end(self, session):
    self._global_step = session.run(self._global_step_tensor)
    self.log()

  def log(self, run_values=None):
    step = self._global_step if self._is_training else self._steps_run_so_far
    if step % self._log_every != 0:
      return
    msg = "{:}/{:} = {:.1f}%".format(step, self._n_steps,
                                     100.0 * step / self._n_steps)
    time_elapsed = time.time() - self._start_time
    time_per_step = time_elapsed / (
        (step - self._start_step) if self._is_training else step)
    msg += ", SPS: {:.1f}".format(1 / time_per_step)
    msg += ", ELAP: " + secs_to_str(time_elapsed)
    msg += ", ETA: " + secs_to_str(
        (self._n_steps - step) * time_per_step)
    if run_values is not None:
      for tag, value in run_values.results.items():
        msg += " - " + str(tag) + (": {:.4f}".format(value))
    log(msg)


def secs_to_str(secs):
  s = str(datetime.timedelta(seconds=int(round(secs))))
  s = re.sub("^0:", "", s)
  s = re.sub("^0", "", s)
  s = re.sub("^0:", "", s)
  s = re.sub("^0", "", s)
  return s


def get_bert_config(config):
  """Get model hyperparameters based on a pretraining/finetuning config"""
  if config.model_size == "large":
    args = {"hidden_size": 1024, "num_hidden_layers": 24}
  elif config.model_size == "base":
    args = {"hidden_size": 768, "num_hidden_layers": 12}
  elif config.model_size == "small":
    args = {"hidden_size": 256, "num_hidden_layers": 12}
  else:
    raise ValueError("Unknown model size", config.model_size)
  args["vocab_size"] = config.vocab_size
  args.update(**config.model_hparam_overrides)
  # by default the ff size and num attn heads are determined by the hidden size
  args["num_attention_heads"] = max(1, args["hidden_size"] // 64)
  args["intermediate_size"] = 4 * args["hidden_size"]
  args.update(**config.model_hparam_overrides)
  return modeling.BertConfig.from_dict(args)


if __name__ == "__main__":
  print("A package for text classifcation using ELECTRA")



# Copyright 2020 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifer: Apache-2.0

import numpy as np
from itertools import product
import matplotlib.pyplot as plt
from sklearn.metrics import accuracy_score, confusion_matrix
from sklearn.metrics import precision_score, recall_score, f1_score


def plot_confusion_matrix(cm, classes = [0, 1], cmap=plt.cm.Accent):

    """ function to plot confusion matrix """

    title = 'Confusion matrix'
    plt.imshow(cm, interpolation='nearest', cmap=cmap)
    plt.title(title)
    plt.colorbar()
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=45)
    plt.yticks(tick_marks, classes)

    fmt = 'd'
    thresh = cm.max() / 2.
    for i, j in product(range(cm.shape[0]), range(cm.shape[1])):
        plt.text(j, i, format(cm[i, j], fmt),
                 horizontalalignment="center",
                 color="black")

    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label')

    
    
def metrics(y, y_pred, average='macro'):
    
    """
    uses sklearn metrics to calculate accuracy, precision, recall, f1_score.

    Parameters
    ----------
    y : actual labels. array-like of shape (n_samples,)
    y_pred : predicted labels. array-like of shape (n_samples,)
    average: determines the type of averaging performed on the data. default: macro
    """

    cm = confusion_matrix(y, y_pred)
    plot_confusion_matrix(cm, np.arange(cm.shape[0]))
    print("\n")
    print("Accuracy: {0:.3f}".format(accuracy_score(y, y_pred)))
    print("Precision: {0:.3f}".format(precision_score(y, y_pred, average=average)))
    print("Recall: {0:.3f}".format(recall_score(y, y_pred, average=average)))
    print("F1-Score: {0:.3f}".format(f1_score(y, y_pred, average=average)))


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()
    
with open("requirements.txt") as fh:
    required = fh.read().splitlines()

setuptools.setup(
    name="classitransformers",
    packages=['classitransformers','electra'],
    version="0.0.1",
    author="FMR LLC",
    author_email="classitransformers@fmr.com",
    description="An abstract library for implementing text classification tasks based on various transformers based language models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fidelity/classitransformers",
    install_requires=required,
    python_requires='>=3.6',
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    keywords='NLP language_models text_classification bert electra roberta distilbert albert',
    
    project_urls={
        "Source": "https://github.com/fidelity/classitransformers"
    }
)




# Copyright 2020 FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifer: Apache-2.0

import os
import sys
import time
import torch
import scipy
import random
import logging
import datetime
import numpy as np
import pandas as pd
from torch.utils.data import TensorDataset, random_split
from transformers import AdamW, get_linear_schedule_with_warmup
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler
from transformers import DistilBertTokenizer, AlbertTokenizer, RobertaTokenizer
from transformers import DistilBertForSequenceClassification, AlbertForSequenceClassification, RobertaForSequenceClassification

logging.getLogger("transformers.tokenization_utils").setLevel(logging.ERROR)


class TransformersClassification:

    def __init__(self, config):
        # self.name, self.num_classes, epochs, batchs
        
        self.Configs = config
        self.num_classes = len(config.label_list)
        
        self.train_logits = []
        self.validation_logits = []
        self.test_logits = []

        self.train_texts = []
        self.train_labels = []
        self.validation_texts = []
        self.validation_labels = []
        self.test_texts = []
        self.test_labels = []

        train = pd.read_csv(os.path.join(self.Configs.data_dir, 'train.csv'))
        
        try:
            dev = pd.read_csv(os.path.join(self.Configs.data_dir, 'dev.csv'))
        
        except:
            print('Validation disabled.')
        test = pd.read_csv(os.path.join(self.Configs.data_dir, 'test.csv'))
        
        self.train_texts = train['text'].tolist()
        
        self.train_labels = train['label'].tolist()
        
        try:
            self.validation_texts = dev['text'].tolist()
            self.validation_labels = dev['label'].tolist()

        except:
            pass
        self.test_texts = test['text'].tolist()
        
        for i in range(len(self.test_texts)):
            self.test_labels.append(0)

        
        if torch.cuda.is_available():
            self.device = torch.device("cuda")
        else:
            print('No GPU available, using the CPU instead.')
            self.device = torch.device("cpu")
        
        if self.Configs.model_name == 'albert':
            self.model = AlbertForSequenceClassification.from_pretrained(self.Configs.pretrained_model_dir,
                                                                         num_labels=self.num_classes)
            self.tokenizer = AlbertTokenizer.from_pretrained(self.Configs.pretrained_model_dir)
        
        if self.Configs.model_name == 'distilbert':
            self.model = DistilBertForSequenceClassification.from_pretrained(self.Configs.pretrained_model_dir,
                                                                             num_labels=self.num_classes)
            self.tokenizer = DistilBertTokenizer.from_pretrained(self.Configs.pretrained_model_dir)
           
        if self.Configs.model_name == 'roberta':
            self.model = RobertaForSequenceClassification.from_pretrained(self.Configs.pretrained_model_dir,
                                                                          num_labels=self.num_classes)
            self.tokenizer = RobertaTokenizer.from_pretrained(self.Configs.pretrained_model_dir)
            
        if torch.cuda.is_available():
            self.model.cuda()
    

    def train(self):

        # Combine the training inputs into a TensorDataset.
        train_params = tokenize_sentences(self.train_texts,self.Configs.max_seq_length,self.tokenizer,self.train_labels)
        train_dataset = TensorDataset(train_params['input_ids'], train_params['attention_masks'],
                                      train_params['labels'])

        if self.Configs.do_eval == True:
            
            validation_params = tokenize_sentences(self.validation_texts, 
                                                       self.Configs.max_seq_length, self.tokenizer, self.validation_labels)
            validation_dataset = TensorDataset(validation_params['input_ids'], validation_params['attention_masks'],
                                               validation_params['labels'])

    
        # Create the DataLoaders for our training and validation sets.
        # We'll take training samples in random order.
        train_dataloader = DataLoader(
            train_dataset,  # The training samples.
            sampler=RandomSampler(train_dataset),  # Select batches randomly
            batch_size=self.Configs.train_batch_size  # Trains with this batch size.
        )

        if self.Configs.do_eval == True:
            validation_dataloader = DataLoader(
                validation_dataset,  # The training samples.
                sampler=RandomSampler(validation_dataset),  # Select batches randomly
                batch_size=self.Configs.eval_batch_size # Trains with this batch size.
            )

        # Note: AdamW is a class from the huggingface library (as opposed to pytorch)
        # I believe the 'W' stands for 'Weight Decay fix"
        optimizer = AdamW(self.model.parameters(),
                          lr=self.Configs.learning_rate,  # default is 5e-5
                          eps=1e-8  # args.adam_epsilon  - default is 1e-8.
                          )

 
        epochs = self.Configs.num_train_epochs

        # Total number of training steps is [number of batches] x [number of epochs].
        # (Note that this is not the same as the number of training samples).
        total_steps = len(train_dataloader) * epochs

        # Create the learning rate scheduler.
        scheduler = get_linear_schedule_with_warmup(optimizer,
                                                    num_warmup_steps=total_steps / 10,  # Default value in run_glue.py
                                                    num_training_steps=total_steps)

  
        # Set the seed value all over the place to make this reproducible.
        seed_val = 42

        random.seed(seed_val)
        np.random.seed(seed_val)
        torch.manual_seed(seed_val)
        torch.cuda.manual_seed_all(seed_val)

        # We'll store a number of quantities such as training and validation loss,
        # validation accuracy, and timings.
        training_stats = []

        # Measure the total training time for the whole run.
        total_t0 = time.time()

        # For each epoch...
        for epoch_i in range(0, epochs):

            # Perform one full pass over the training set.

            print("")
            print('======== Epoch {:} / {:} ========'.format(epoch_i + 1, epochs))
            print('Training...')

            # Measure how long the training epoch takes.
            t0 = time.time()

            # Reset the total loss for this epoch.
            total_train_loss = 0
            train_logits = []
            validation_logits = []
            
            self.model.train()

            # For each batch of training data...
            for step, batch in enumerate(train_dataloader):

                # Progress update every 40 batches.
                if step % 40 == 0 and not step == 0:
                    # Calculate elapsed time in minutes.
                    elapsed = format_time(time.time() - t0)

                    # Report progress.
                    print(
                        '  Batch {:>5,}  of  {:>5,}.    Elapsed: {:}.'.format(step, len(train_dataloader), elapsed))

                b_input_ids = batch[0].to(self.device)
                b_input_mask = batch[1].to(self.device)
                b_labels = batch[2].to(self.device)

                self.model.zero_grad()

                loss, logits = self.model(b_input_ids,

                                          attention_mask=b_input_mask,
                                          labels=b_labels)

                logits = logits.detach().cpu().numpy()
                label_ids = b_labels.to('cpu').numpy()
                train_logits.append(logits)

                # Accumulate the training loss over all of the batches so that we can
                # calculate the average loss at the end. 
                total_train_loss += loss.item()

                # Perform a backward pass to calculate the gradients.
                loss.backward()

                # Clip the norm of the gradients to 1.0.
                # This is to help prevent the "exploding gradients" problem.
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), 1.0)

                # Update parameters and take a step using the computed gradient.
                optimizer.step()

                # Update the learning rate.
                scheduler.step()

            # Calculate the average loss over all of the batches.
            avg_train_loss = total_train_loss / len(train_dataloader)

            # Measure how long this epoch took.
            training_time = format_time(time.time() - t0)

            print("")
            print("  Average training loss: {0:.2f}".format(avg_train_loss))
            print("  Training epcoh took: {:}".format(training_time))

            torch.cuda.empty_cache()

            if self.Configs.do_eval == True:
                
                # After the completion of each training epoch, measure our performance on
                # our validation set.

                print("")
                print("Running Validation...")

                t0 = time.time()

                # Put the model in evaluation mode--the dropout layers behave differently
                # during evaluation.
                self.model.eval()

                # Tracking variables
                total_eval_accuracy = 0
                total_eval_loss = 0
                nb_eval_steps = 0

                # Evaluate data for one epoch
                for batch in validation_dataloader:
                    
                    b_input_ids = batch[0].to(self.device)
                    b_input_mask = batch[1].to(self.device)
                    b_labels = batch[2].to(self.device)

                    # Tell pytorch not to bother with constructing the compute graph during
                    # the forward pass, since this is only needed for backprop (training).
                    with torch.no_grad():
                        
                        loss, logits = self.model(b_input_ids,
                                                  attention_mask=b_input_mask,
                                                  labels=b_labels)

                        # Move logits and labels to CPU
                        logits = logits.detach().cpu().numpy()
                        label_ids = b_labels.to('cpu').numpy()
                        validation_logits.append(logits)
                    # Accumulate the validation loss.
                    # total_eval_loss += loss.item()

                    # Calculate the accuracy for this batch of test sentences, and
                    # accumulate it over all batches.
                    total_eval_accuracy += flat_accuracy(logits, label_ids)

                # Report the final accuracy for this validation run.
                avg_val_accuracy = total_eval_accuracy / len(validation_dataloader)
                print("  Accuracy: {0:.2f}".format(avg_val_accuracy))

                # Calculate the average loss over all of the batches.
                avg_val_loss = total_eval_loss / len(validation_dataloader)

                # Measure how long the validation run took.
                validation_time = format_time(time.time() - t0)

                print("  Validation Loss: {0:.2f}".format(avg_val_loss))
                print("  Validation took: {:}".format(validation_time))


            # torch.cuda.empty_cache()

            # Record all statistics from this epoch.
            training_stats.append(
                {
                    'epoch': epoch_i + 1,
                    'Training Loss': avg_train_loss,
                    'Valid. Loss': avg_val_loss if self.Configs.do_eval == True else 'NA',
                    'Valid. Accur.': avg_val_accuracy if self.Configs.do_eval == True else 'NA',
                    'Training Time': training_time,
                    'Validation Time': validation_time if self.Configs.do_eval == True else 'NA'
                }
            )
        
        if self.Configs.do_eval == True:
            validation_logits = np.vstack(validation_logits)
        
        train_logits = np.vstack(train_logits)

        self.train_logits = train_logits
        self.validation_logits = validation_logits

        # torch.state_dict(self.model,'/content/checkpoint.pth')
        output_dir = self.Configs.output_dir

        # Create output directory if needed
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        print("Saving model to %s" % output_dir)

        # Save a trained model, configuration and tokenizer using `save_pretrained()`.
        # They can then be reloaded using `from_pretrained()`
        model_to_save = self.model.module if hasattr(self.model, 'module') else self.model  # Take care of distributed/parallel training
        model_to_save.save_pretrained(output_dir)
        self.tokenizer.save_pretrained(output_dir)

        print("")
        print("Training complete!")

        print("Total training took {:} (h:mm:ss)".format(format_time(time.time() - total_t0)))


    def test(self):

        test_params = tokenize_sentences(self.test_texts,self.Configs.max_seq_length,self.tokenizer,self.test_labels)
        
        prediction_data = TensorDataset(test_params['input_ids'], test_params['attention_masks'], test_params['labels'])
        prediction_sampler = SequentialSampler(prediction_data)
        prediction_dataloader = DataLoader(prediction_data, sampler=prediction_sampler, batch_size=self.Configs.predict_batch_size)

        
        # Put model in evaluation mode
        self.model.eval()

        # Tracking variables
        predictions = []

        # Predict
        for batch in prediction_dataloader:
            # Add batch to GPU
            # Unpack the inputs from our dataloader
            b_input_ids, b_input_mask, b_labels = tuple(t.to(self.device) for t in batch)

            # Telling the model not to compute or store gradients, saving memory and
            # speeding up prediction
            with torch.no_grad():
                # Forward pass, calculate logit predictions
                loss, logits = self.model(b_input_ids,
                                          attention_mask=b_input_mask,
                                          labels=b_labels)

            # Move logits and labels to CPU
            logits = logits.detach().cpu().numpy()

            # Store predictions and true labels
            predictions.extend(logits)
         
        self.test_logits = predictions
        torch.cuda.empty_cache()

        return [scipy.special.expit(pred) for pred in predictions]
    
    
    def text_inference(self, texts):

        inf_labels = []
        for i in range(len(texts)):
            inf_labels.append(0)

        test_params = tokenize_sentences(texts, self.Configs.max_seq_length, self.tokenizer, inf_labels)
        prediction_data = TensorDataset(test_params['input_ids'], test_params['attention_masks'],
                                        test_params['labels'])
        prediction_sampler = SequentialSampler(prediction_data)
        prediction_dataloader = DataLoader(prediction_data, sampler=prediction_sampler,
                                           batch_size=self.Configs.predict_batch_size)

        self.model.eval()
        predictions = []
        
        for batch in prediction_dataloader:
            b_input_ids, b_input_mask, b_labels = tuple(t.to(self.device) for t in batch)
            with torch.no_grad():
                loss, logits = self.model(b_input_ids,
                                          attention_mask=b_input_mask,
                                          labels=b_labels)

            logits = logits.detach().cpu().numpy()
            predictions.extend(logits)
           
        self.test_logits = predictions
        torch.cuda.empty_cache()
        return [scipy.special.expit(pred) for pred in predictions]


def add_CLS_and_SEP(sentences,tokenizer):
    max_len = 0

    for sent in sentences:
        # Tokenize the text and add `[CLS]` and `[SEP]` tokens.
        input_ids = tokenizer.encode(sent, add_special_tokens=True)

        # Update the maximum sentence length.
        max_len = max(max_len, len(input_ids))

    
    return sentences


def tokenize_sentences(sentences,max_seq_len,tokenizer,labels = []):
    # Tokenize all of the sentences and map the tokens to thier word IDs.
    input_ids = []
    attention_masks = []

    sentences = add_CLS_and_SEP(sentences,tokenizer)

    for sent in sentences:
        
        encoded_dict = tokenizer.encode_plus(
            sent,  # Sentence to encode.
            add_special_tokens=True,  # Add '[CLS]' and '[SEP]'
            max_length=max_seq_len,  # Pad & truncate all sentences.
            pad_to_max_length=True,
            return_attention_mask=True,  # Construct attn. masks.
            return_tensors='pt',  # Return pytorch tensors.
        )

        # Add the encoded sentence to the list.
        input_ids.append(encoded_dict['input_ids'])

        # And its attention mask (simply differentiates padding from non-padding).
        attention_masks.append(encoded_dict['attention_mask'])

    # Convert the lists into tensors.
    input_ids = torch.cat(input_ids, dim=0)
    attention_masks = torch.cat(attention_masks, dim=0)
    try:
        labels = torch.tensor(labels)
    except:
        labels = []
        for i in range(len(sentences)):
            labels.append[0]

    return {
        'input_ids': input_ids,
        'attention_masks': attention_masks,
        'labels': labels
    }



# Function to calculate the accuracy of our predictions vs labels
def flat_accuracy(preds, labels):
    pred_flat = np.argmax(preds, axis=1).flatten()
    labels_flat = labels.flatten()
    return np.sum(pred_flat == labels_flat) / len(labels_flat)




def format_time(elapsed):
    '''
    Takes a time in seconds and returns a string hh:mm:ss
    '''
    # Round to the nearest second.
    elapsed_rounded = int(round((elapsed)))

    # Format as hh:mm:ss
    return str(datetime.timedelta(seconds=elapsed_rounded))